#!/usr/bin/env python
# (C) Copyright 2021 JSC.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, JSC does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
"""
Class to build up a Unet for downsclaing following Sha et al., 2020 (DOI:
"""
import numpy as np
import xarray as xr
from typing import Union, List

import tensorflow as tf
import tensorflow.keras as keras

# all the layers used for U-net
from tensorflow.keras.layers import (Activation, BatchNormalization, Concatenate, Conv2D, Conv2DTranspose, Input,
                                     MaxPool2D)
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import backend as K

List_or_tuple = Union[List, tuple]

class Unet_Downscaling(object):

    class_name = "Unet_Downscaling"

    def __init__(self, input_shape: List_or_tuple, model_name: str, base_loss="mae", channels_start: int = 56,
                 z_branch: bool = False, sobel_loss: bool = False):

        method = "{0}->{1}".format(Unet_Downscaling.class_name, Unet_Downscaling.__init__.__name__)

        # sanity checks of passed arguments
        assert isinstance(input_shape, List_or_tuple), "%{0}: Passed input_shape must be a list or tuple, but is {1}"\
                                                       .format(method, type(input_shape))
        assert len(input_shape) == 3, "%{0} Length of input_shape must be three, but is {1:d}".format(method,
                                                                                                      len(input_shape))
        assert isinstance(model_name, str), "%{0}: Passed model_name must be a string".format(method)
        assert int(channels_start) >= 4, "%{0}: channels_start must be at least 4.".format(method)
        assert isinstance(z_branch, bool), "%{0}: z_branch must a boolean, but is of type {1}".format(method,
                                                                                                      type(z_branch))
        assert not sobel_loss, "%{0}: sobel_loss is not implemented yet and must therefore be False.".format(method)

        self.input_shape = input_shape
        self.model_name = model_name
        self.base_loss = Unet_Downscaling.get_keras_loss(base_loss)
        self.channels_start = channels_start
        self.build_z_branch = z_branch
        self.apply_sobel_loss = sobel_loss

        self.l_compile = False
        self.l_train = False
        self.callback = None        # set in compile_unet method
        self.batch_size = None      # set in train_unet method

        self.unet_model = self.build_unet()

    def build_unet(self):
        """
        Builds the U-net
        :return: built model with Keras API
        """
        method = "{0}->{1}".format(Unet_Downscaling.class_name, Unet_Downscaling.build_unet.__name__)

        inputs = Input(self.input_shape)
        channels_start = self.channels_start

        """ encoder """
        s1, e1 = Unet_Downscaling.encoder_block(inputs, channels_start, l_large=True)
        s2, e2 = Unet_Downscaling.encoder_block(e1, channels_start * 2, l_large=False)
        s3, e3 = Unet_Downscaling.encoder_block(e2, channels_start * 4, l_large=False)

        """ bridge encoder <-> decoder """
        b1 = Unet_Downscaling.conv_block(e3, channels_start * 8)

        """ decoder """
        d1 = Unet_Downscaling.decoder_block(b1, s3, channels_start * 4)
        d2 = Unet_Downscaling.decoder_block(d1, s2, channels_start * 2)
        d3 = Unet_Downscaling.decoder_block(d2, s1, channels_start)

        output_temp = Conv2D(1, (1, 1), kernel_initializer="he_normal", name="output_temp")(d3)
        if self.build_z_branch:
            output_z = Conv2D(1, (1, 1), kernel_initializer="he_normal", name="output_z")(d3)

            model = Model(inputs, [output_temp, output_z], name=self.model_name)
        else:
            model = Model(inputs, output_temp, name=self.model_name)

        return model

    def compile_unet(self, lr_scheduler: callable= None):
        """
        Compiles previously built U-net
        :param lr_scheduler: learning reate scheduler
        :return:
        """
        method = "{0}->{1}".format(Unet_Downscaling.class_name, Unet_Downscaling.compile_unet.__name__)

        if lr_scheduler is None:
            def lr_scheduler(epoch, lr):
                if epoch < 5:
                    return lr
                elif 5 <= epoch < 30:
                    return lr * tf.math.exp(-0.1)
                elif epoch >= 30:
                    return lr

        try:
            self.callback = tf.keras.callbacks.LearningRateScheduler(lr_scheduler)
        except Exception as err:
            print("%{0}: Failed to built learning rate scheduler. See risen exception.".format(method))
            raise err

        if self.build_z_branch:
            self.unet_model.compile(optimizer=Adam(learning_rate=5 * 10 ** (-4)),
                                    loss={"output_temp": self.base_loss,
                                          "output_z": self.base_loss},
                                    loss_weights={"output_temp": 1.0, "output_z": 1.0}, metrics=["mse"])
        else:
            self.unet_model.compile(optimizer=Adam(learning_rate=5 * 10 ** (-4)), loss=self.base_loss, metrics=["mse"])

        self.l_compile = True
        self.l_train = False

    def train_unet(self, in_data: xr.DataArray, tar_data: xr.DataArray, batch_size=16, epochs: int = 50):
        """
        Train the U-net.
        :param in_data: input data for training
        :param tar_data: target data for supervised training
        :param batch_size: number of training samples per mini-batch
        :param epochs: number of epocs for training
        :return: history of model training (from Keras API)
        """
        method = "{0}->{1}".format(Unet_Downscaling.class_name, Unet_Downscaling.train_unet.__name__)

        # sanity checks
        assert isinstance(in_data, xr.DataArray), "%{0}: in_data must be a xarray DataArray.".format(method)
        assert isinstance(tar_data, xr.DataArray), "%{0}: tar_data must be a xarray DataArray.".format(method)

        if not self.l_compile:
            raise RuntimeError("%{0}: Unet-model has not been compiled yet.".format(method))

        self.batch_size = batch_size
        # Fit the model
        if self.build_z_branch:
            history_model = self.unet_model.fit(x=in_data.values, y={"output_temp": tar_data.isel(variable=0).values,
                                                                     "output_z": tar_data.isel(variable=1).values},
                                                epochs=epochs, batch_size=batch_size, callbacks=[self.callback])
        else:
            history_model = self.unet_model.fit(x=in_data.values, y=tar_data.isel(variable=0).values, epochs=epochs,
                                                batch_size=batch_size, callbacks=[self.callback])
        # change training status
        self.l_train = True

        return history_model

    def predict(self, in_data: xr.DataArray):
        """
        Create predictions from trained model.
        :param in_data: input data for creating predictions
        :return: predictions from U-net
        """
        method = "{0}->{1}".format(Unet_Downscaling.class_name, Unet_Downscaling.predict.__name__)

        # sanity checks
        assert isinstance(in_data, xr.DataArray), "%{0}: in_data must be a xarray DataArray.".format(method)

        if not self.l_train:
            raise RuntimeError("%{0}: Model {1} is still untrained. Cannot create predictions."
                               .format(method, self.model_name))

        y_pred = self.unet_model.predict(in_data.values, verbose=1)

        return y_pred

    def plot_unet_model(self, filename: str = None):
        """
        Creates plot of U-net
        :param filename: filename to save plot
        :return: -
        """
        method = "{0}->{1}".format(Unet_Downscaling.class_name, Unet_Downscaling.plot_unet_model.__name__)

        if filename is None:
            filename = "plot_{0}".format(self.model_name)
            print("%{0}: Defaulting filename of model plot to {1}".format(method, filename))

        try:
            keras.utils.plot_model(self.unet_model, show_shapes=True, to_file=filename+".png")
        except Exception as err:
            print("%{0}: Failed to plot model. See risen error message.".format(method))
            raise err

    def get_model_memory_usage(self, model=None):
        """
        Calculates memory size of model in gigabyte
        :param model: model instance
        :return: size of memory for U-net model
        """
        method = "{0}->{1}".format(Unet_Downscaling.class_name, Unet_Downscaling.get_model_memory_usage.__name__)

        if model is None:
            model = self.unet_model
        shapes_mem_count = 0
        internal_model_mem_count = 0
        for l in model.layers:
            layer_type = l.__class__.__name__
            if layer_type == 'Model':
                internal_model_mem_count += self.get_model_memory_usage(model=l)
            single_layer_mem = 1
            out_shape = l.output_shape
            if type(out_shape) is list:
                out_shape = out_shape[0]
            for s in out_shape:
                if s is None:
                    continue
                single_layer_mem *= s
            shapes_mem_count += single_layer_mem

        trainable_count = np.sum([K.count_params(p) for p in model.trainable_weights])
        non_trainable_count = np.sum([K.count_params(p) for p in model.non_trainable_weights])

        number_size = 4.0
        if K.floatx() == 'float16':
            number_size = 2.0
        if K.floatx() == 'float64':
            number_size = 8.0

        total_memory = number_size * (self.batch_size * shapes_mem_count + trainable_count + non_trainable_count)
        gbytes = np.round(total_memory / (1024.0 ** 3), 3) + internal_model_mem_count
        return gbytes

    @staticmethod
    def get_keras_loss(loss_name: str):
        """
        Sets pointer to Keras loss.
        :param loss_name: name of loss, must be either "mae", "mape", "mse" or "msle"
        :return: loss-function
        """
        method = Unet_Downscaling.get_keras_loss.__name__

        loss_name = loss_name.lower()

        if loss_name == "mae":
            loss = keras.losses.MeanAbsoluteError()
        elif loss_name == "mape":
            loss = keras.losses.MeanAbsolutePercentageError()
        elif loss_name == "mse":
            loss = keras.losses.MeanSquaredError()
        elif loss_name == "msle":
            loss = keras.losses.MeanSquaredLogarithmicError()
        else:
            raise ValueError("%{0}: Passed loss_name {1} is unknown".format(method, loss_name))

        return loss

    @staticmethod
    def conv_block(inputs, num_filters: int, kernel: tuple = (3, 3), padding: str = "same",
                   activation: str = "relu", kernel_init: str = "he_normal", l_batch_normalization: bool = True):
        """
        A convolutional layer with optional batch normalization
        :param inputs: the input data with dimensions nx, ny and nc
        :param num_filters: number of filters (output channel dimension)
        :param kernel: tuple indictating kernel size
        :param padding: technique for padding (e.g. "same" or "valid")
        :param activation: activation fuction for neurons (e.g. "relu")
        :param kernel_init: initialization technique (e.g. "he_normal" or "glorot_uniform")
        :param l_batch_normalization: Boolean if batch normalization should take place
        :return: x from inputs processed through convolutional block
        """
        x = Conv2D(num_filters, kernel, padding=padding, kernel_initializer=kernel_init)(inputs)
        if l_batch_normalization:
            x = BatchNormalization()(x)
        x = Activation(activation)(x)

        return x

    @staticmethod
    def conv_block_n(inputs, num_filters, n=2, kernel=(3, 3), padding="same", activation="relu",
                     kernel_init="he_normal", l_batch_normalization=True):
        """
        Sequential application of two convolutional layers (using conv_block).
        """

        x = Unet_Downscaling.conv_block(inputs, num_filters, kernel, padding, activation,
                                        kernel_init, l_batch_normalization)
        for i in np.arange(n - 1):
            x = Unet_Downscaling.conv_block(x, num_filters, kernel, padding, activation,
                                            kernel_init, l_batch_normalization)

        return x

    @staticmethod
    def encoder_block(inputs, num_filters, kernel_maxpool: tuple = (2, 2), l_large: bool = True):
        """
        One complete encoder-block used in U-net
        """
        if l_large:
            x = Unet_Downscaling.conv_block_n(inputs, num_filters, n=2)
        else:
            x = Unet_Downscaling.conv_block(inputs, num_filters)

        p = MaxPool2D(kernel_maxpool)(x)

        return x, p

    @staticmethod
    def decoder_block(inputs, skip_features, num_filters, kernel: tuple = (3, 3), strides_up: int = 2,
                      padding: str = "same", activation="relu", kernel_init="he_normal",
                      l_batch_normalization: bool = True):
        """
        One complete decoder block used in U-net (reverting the encoder)
        """
        x = Conv2DTranspose(num_filters, (strides_up, strides_up), strides=strides_up, padding="same")(inputs)
        x = Concatenate()([x, skip_features])
        x = Unet_Downscaling.conv_block_n(x, num_filters, 2, kernel, padding, activation, kernel_init,
                                          l_batch_normalization)

        return x


